{*******************************************************}
{                                                       }
{  BallDetect                                           }
{  Copyright (c) 2003 by Andreas Wulf                   }
{                                                       }
{*******************************************************}
{                                                       }
{  Provides a function (FindRedBall) that attempts to   }
{  detect a red ball in a given image.                  }
{                                                       }
{  Last modified: Oct 31, 2003                          }
{                                                       }
{*******************************************************}

unit BallDetect;

interface

uses
  Windows, Classes, Graphics, Math, Forms;

type
  TBallData = record
    XPos: Integer;
    YPos: Integer;
    Radius: Integer;
  end;

  TCandidate = class
    MinX, MaxX, MinY, MaxY: Integer;
    MinR, MaxR: Integer;
  end;

  THLS = record
    H: Integer;  // Hue
    L: Integer;  // Lightness
    S: Integer;  // Saturation
  end;

  PRGBArray = ^TRGBArray;
  TRGBArray = array [0..319] of TRGBTriple;
  PRGBRows = ^TRGBRows;
  TRGBRows = array[0..239] of PRGBArray;
  TImageData = record
    Rows: PRGBRows;
    Width: Integer;
    Height: Integer;
  end;

function FindRedBall(Bitmap: TBitmap): TBallData;  // Bitmap size must be 320x240 pixels

implementation

var
  CandidateList: TList;  // list of candidates where circles are supposed to be located

const
  // allowed image size
  LargeWidth = 320;
  LargeHeight = 240;

  // image will be resampled to the following dimensions in order to guarantee
  // a reasonable performance
  SmallWidth = 128;
  SmallHeight = 96;

{ *** Resample code *** }

function TrimReal(Lower, Upper: Integer; X: Double): Integer;
begin
  if (X < Upper) and (X >= Lower) then
    Result:= Trunc(X)
  else
    if X >= Upper then
      Result:= Upper
    else
      Result:= Lower;
end;

// Mitchell resample filter
function MitchellFilter(Value: Single): Single;
const
  B = (1.0 / 3.0);
  C = (1.0 / 3.0);
var
  TT: single;
begin
  if (Value < 0.0) then
    Value := -Value;
  TT := Sqr(Value);
  if (Value < 1.0) then
  begin
    Value := (((12.0 - 9.0 * B - 6.0 * C) * (Value * TT))
      + ((-18.0 + 12.0 * B + 6.0 * C) * TT)
      + (6.0 - 2 * B));
    Result := Value / 6.0;
  end else
  if (Value < 2.0) then
  begin
    Value := (((-1.0 * B - 6.0 * C) * (Value * TT))
      + ((6.0 * B + 30.0 * C) * TT)
      + ((-12.0 * B - 48.0 * C) * Value)
      + (8.0 * B + 24 * C));
    Result := Value / 6.0;
  end else
    Result := 0.0;
end;

// Resample bitmap to given extents
// based on Anders Melander's "Interpolating Bitmap Resampler"
// http://www.melander.dk/delphi/resampler/index.html
procedure Resample(Bitmap: TBitmap; NewWidth, NewHeight: Integer);
type
  TColorRGB = packed record
    R, G, B: Byte;
  end;
  PColorRGB = ^TColorRGB;

  TContributor = record
    Pixel: integer;
    Weight: single;
  end;

  TContributorList = Array[0..0] of TContributor;
  PContributorList = ^TContributorList;

  TCList = record
    n: integer;
    p: PContributorList;
  end;

  TCListList = Array[0..0] of TCList;
  PCListList = ^TCListList;

  TRGB = packed record
    R, G, B: Single;
  end;
  TRGBList = packed Array[0..0] of TColorRGB;
  PRGBList = ^TRGBList;
var
  XScale, YScale: Single;         // Zoom scale factors
  I, J, K: Integer;               // Loop variables
  Center: Single;                 // Filter calculation variables
  Width, FScale, Weight: Single;  // Filter calculation variables
  Left, Right: Integer;           // Filter calculation variables
  N: Integer;                     // Pixel number
  Work: TBitmap;                  // Temporary Bitmap
  Contrib: PCListList;            // Contributor pointer
  RGB: TRGB;                      // RGBTriple
  Color: TColorRGB;               // ColorRGBTriple
  SourceLine, DestLine: PRGBList;
  SourcePixel, DestPixel: PColorRGB;
  Delta, DestDelta: Integer;
  SMode: Boolean;
begin
  if (Bitmap.Width < 1) or (Bitmap.Height < 1) then
    Exit;

  Application.ProcessMessages;

  Work := TBitmap.Create;
  try
    Work.Height := Bitmap.Height;
    Work.Width := NewWidth;
    if (Bitmap.Width = 1) then
      XScale := NewWidth / Bitmap.Width
    else
      XScale := (NewWidth - 1) / (Bitmap.Width - 1);
    if (Bitmap.Height = 1) then
      YScale := NewHeight / Bitmap.Height
    else
      YScale := (NewHeight - 1) / (Bitmap.Height - 1);

    Bitmap.PixelFormat := pf24bit;
    Work.PixelFormat := Bitmap.PixelFormat;

    { Horizontal Part }

    // Pre-calculate filter contributions for a row
    GetMem(Contrib, NewWidth * SizeOf(TCList));

    // Horizontal sub-/supersampling

    // set different modes depending on scaling factor
    SMode := XScale < 1.0;
    if SMode then
      Width := 3 / XScale
    else
      Width := 3;
    if SMode then
      FScale := 1.0 / XScale
    else
      FScale:=1;

    for I := 0 to NewWidth - 1 do
    begin
      Contrib^[I].N := 0;
      GetMem(contrib^[I].P, Trunc(Width * 2.0 + 1) * SizeOf(TContributor));
      Center := I / XScale;
      Left := Floor(Center - Width);
      Right := Ceil(Center + Width);
      for J := Left to Right do
      begin
        Weight := MitchellFilter((Center - j) / FScale) / FScale;
        if (Weight = 0.0) then
          Continue;
        if (J < 0) then
          N := -J
        else if (J >= Bitmap.Width) then
          N := Bitmap.Width - J + Bitmap.Width - 1
        else
          N := J;
        K := Contrib^[I].N;
        Contrib^[I].N := Contrib^[i].N + 1;
        Contrib^[I].P^[k].Pixel := N;
        Contrib^[I].P^[k].Weight := Weight;
      end;
    end;

    // Apply filter to sample horizontally from Src to Work
    for K := 0 to Bitmap.Height - 1 do
    begin
      SourceLine := Bitmap.ScanLine[K];
      DestPixel := Work.ScanLine[K];
      for I := 0 to NewWidth - 1 do
      begin
        RGB.R := 0.0;
        RGB.G := 0.0;
        RGB.B := 0.0;
        for J := 0 to Contrib^[I].N-1 do
        begin
          Color := SourceLine^[Contrib^[I].P^[J].Pixel];
          Weight := Contrib^[I].P^[J].weight;
          if (Weight = 0.0) then
            Continue;
          RGB.R := RGB.R + Color.R * Weight;
          RGB.G := RGB.G + Color.G * Weight;
          RGB.B := RGB.B + Color.B * Weight;
        end;
        Color.R := TrimReal(0, 255, RGB.R);
        Color.G := TrimReal(0, 255, RGB.G);
        Color.B := TrimReal(0, 255, RGB.B);
        // Set new pixel value
        DestPixel^ := Color;
        // Move on to next column
        inc(DestPixel);
      end;
      Application.ProcessMessages;
    end;

    // Free the memory (horizontal filter weights)
    for I := 0 to NewWidth - 1 do
      FreeMem(Contrib^[I].P);

    FreeMem(Contrib);

    { Vertical part }

    // Pre-calculate filter contributions for a column
    GetMem(Contrib, NewHeight * SizeOf(TCList));

    // Vertical sub-/supersampling

    SMode := YScale < 1.0;
    if SMode then
      Width := 3 / YScale
    else
      Width := 3;
    if SMode then
      FScale := 1.0 / YScale
    else
      FScale := 1;
    for I := 0 to NewHeight - 1 do
    begin
      Contrib^[I].N := 0;
      GetMem(Contrib^[I].P, Trunc(Width * 2.0 + 1) * SizeOf(TContributor));
      Center := I / YScale;
      Left := Floor(Center - Width);
      Right := Ceil(Center + Width);
      for J := Left to Right do
      begin
        Weight := MitchellFilter((Center - J) / FScale) / FScale;
        if (Weight = 0.0) then
          Continue;
        if (J < 0) then
          N := -J
        else if (j >= Bitmap.Height) then
          N := Bitmap.Height - J + Bitmap.Height - 1
        else
          N := J;
        K := Contrib^[I].N;
        Contrib^[I].N := Contrib^[I].N + 1;
        Contrib^[I].P^[K].Pixel := N;
        Contrib^[I].P^[K].Weight := Weight;
      end;
    end;

    Bitmap.Width := NewWidth;
    Bitmap.Height := NewHeight;

    // Apply filter to sample vertically from Work to Dst
    SourceLine := Work.ScanLine[0];
    Delta := Integer(Work.ScanLine[1]) - Integer(SourceLine);
    DestLine := Bitmap.ScanLine[0];
    DestDelta := Integer(Bitmap.ScanLine[1]) - Integer(DestLine);
    for K := 0 to NewWidth - 1 do
    begin
      DestPixel := Pointer(DestLine);
      for I := 0 to NewHeight - 1 do
      begin
        RGB.R := 0;
        RGB.G := 0;
        RGB.B := 0;
        for J := 0 to Contrib^[I].N - 1 do
        begin
          Color := PColorRGB(Integer(SourceLine) + Contrib^[I].P^[J].Pixel * Delta)^;
          Weight := Contrib^[I].P^[J].Weight;
          if (Weight = 0.0) then
            Continue;
          RGB.R := RGB.R + Color.R * Weight;
          RGB.G := RGB.G + Color.G * Weight;
          RGB.B := RGB.b + Color.B * Weight;
        end;
        Color.R := TrimReal(0, 255, RGB.R);
        Color.G := TrimReal(0, 255, RGB.G);
        Color.B := TrimReal(0, 255, RGB.B);
        DestPixel^ := Color;
        inc(Integer(DestPixel), DestDelta);
      end;
      Inc(SourceLine, 1);
      Inc(DestLine, 1);
      Application.ProcessMessages;
    end;

    // Free the memory (vertical filter weights)
    for I := 0 to NewHeight - 1 do
      FreeMem(Contrib^[i].P);

    FreeMem(Contrib);

  finally
    Work.Free;
  end;
end;

{ *** Edge finding code *** }

function TrimInt(X, Lower, Upper: Integer): Integer;
begin
  if (X < Upper) and (x >= Lower) then
    Result := X
  else
    if X >= Upper then
      Result := Upper
    else
      Result := Lower;
end;

function GetRGBTriple(ImageData: TImageData; Row, Col: Integer): TRGBTriple;
begin
  if (Row >= 0) and (Row < ImageData.Height) and (Col >= 0) and (Col < ImageData.Width) then
    Result := ImageData.Rows[Row][Col]
  else
    with Result do
    begin
      rgbtBlue := 0;
      rgbtGreen := 0;
      rgbtRed := 0;
    end;
end;

// Find edges in image
// based on code taken from Andreas Moser's "GFX_LIBRARY v1.5"
// http://www.amoser.de/gfxlib_frame.html
procedure EdgeFilter(SrcImgData: TImageData; var DestBitmap: TBitmap);
const
  Matrix: Array [-1..1, -1..1] of Integer =
     ((1, 3, 1),
      (3, -16, 3),
      (1, 3, 1));
var
  Row, Col: Integer;
  I: Integer;
  HorzLine, VertLine: Integer;
  TargetRow: PRGBArray;
  Val1, Val2, Val3: Array [-1..1] of Integer;
  Val1Sum, Val2Sum, Val3Sum: Integer;
  BorderDiffX, BorderDiffY: Integer;
begin
  for Row := 0 to SrcImgData.Height - 1 do
  begin
    TargetRow := DestBitmap.Scanline[Row];
    for Col := 0 to SrcImgData.Width - 1 do
    begin
      Val1Sum := 0;
      Val2Sum := 0;
      Val3Sum := 0;
      for I := -1 to 1 do
      begin
        Val1[I] := 0;
        Val2[I] := 0;
        Val3[I] := 0;
      end;
      for VertLine := -1 to 1 do
      begin
        if ((Row < 1) and (Vertline < 0)) or ((Row > SrcImgData.Height - 2) and (Vertline > 0)) then
          BorderDiffY := -1
        else
          BorderDiffY := 1;
        for HorzLine := -1 to 1 do
        begin
          if ((Col < 1) and (Horzline < 0)) or ((Col > SrcImgData.Width - 2) and (Horzline > 0)) then
            BorderDiffX :=-1
          else
            BorderDiffX := 1;
          Val1[VertLine] := Val1[VertLine] + (GetRGBTriple(SrcImgData, Row + VertLine * BorderDiffY, Col + HorzLine*BorderDiffX).rgbtRed * Matrix[HorzLine, VertLine]);
          Val2[VertLine] := Val2[VertLine] + (GetRGBTriple(SrcImgData, Row + VertLine * BorderDiffY, Col + HorzLine*BorderDiffX).rgbtGreen * Matrix[HorzLine, VertLine]);
          Val3[Vertline] := Val3[VertLine] + (GetRGBTriple(SrcImgData, Row + VertLine * BorderDiffY, Col + HorzLine*BorderDiffX).rgbtBlue * Matrix[HorzLine, VertLine]);
        end;
      end;
      for I := -1 to 1 do
      begin
        Val1Sum := Val1Sum + Val1[I];
        Val2Sum := Val2Sum + Val2[I];
        Val3Sum := Val3Sum + Val3[I];
      end;
      TargetRow[Col].rgbtRed := TrimInt(Val1Sum, 0, 255);
      TargetRow[Col].rgbtGreen := TrimInt(Val2Sum, 0, 255);
      TargetRow[Col].rgbtBlue := TrimInt(Val3Sum, 0, 255);
    end;
    Application.ProcessMessages;
  end;
end;

{ *** Color functions *** }

function RGBToHLS(PRGB: TRGBTriple): THLS;
var
  LR, LG, LB, LH, LL, LS, LMin, LMax: Double;
begin
  LR := PRGB.rgbtRed / 256;
  LG := PRGB.rgbtGreen / 256;
  LB := PRGB.rgbtBlue / 256;
  LMin := MinValue([LR, LG, LB]);
  LMax := MaxValue([LR, LG, LB]);
  LL := (LMax + LMin) / 2;
  if LMin = LMax then
  begin
    Result.H := 0;
    Result.L := Round(LL * 256);
    Result.S := 0;
    exit;
  end;
  if LL < 0.5 then
    LS := (LMax - LMin) / (LMax + LMin)
  else
    LS := (LMax - LMin) / (2.0 - LMax - LMin);
  If LR = LMax then
    LH := (LG - LB) / (LMax - LMin)
  else
    If LG = LMax then
      LH := 2.0 + (LB - LR) / (LMax - LMin)
    else
      LH := 4.0 + (LR - LG) / (LMax - LMin);

  Result.H := Round(LH * 42.6);
  if Result.H < 0 then
    Result.H := Result.H + 255;
  Result.L := Round(LL * 256);
  Result.S := Round(LS * 256);
end;

function IsRedTone(Color: TRGBTriple): Boolean;
var
  HLSColor: THLS;
  AbsHVal: Integer;
begin
  HLSColor := RGBToHLS(Color);

  Result := False;

  AbsHVal := Abs(((HLSColor.H + 14) mod 255) - 14);

  if (AbsHVal > 14) then
    Exit;

  if ((HLSColor.S >= 80 + AbsHVal * 3) and (HLSColor.S <= 256))
    and ((HLSColor.L >= 60 + AbsHVal * 4) and (HLSColor.L <= 210 - AbsHVal * 4)) then
    Result := True;

  {if ((HLSColor.H >= 250) or (HLSColor.H <= 10))
    and ((HLSColor.S >= 80) and (HLSColor.S <= 256))
    and ((HLSColor.L >= 70) and (HLSColor.L <= 200)) then
    Result := True
  else
    Result := False;}
end;

function IsDarkTone(Color: TRGBTriple): Boolean;
var
  HLSColor: THLS;
begin
  HLSColor := RGBToHLS(Color);

  if HLSColor.L <= 100 then
    Result := true
  else
    Result := false;
end;

function IsBlack(OrgColor: TRGBTriple): Boolean;
begin
  if (OrgColor.rgbtRed = 0) and (OrgColor.rgbtGreen = 0) and (OrgColor.rgbtBlue = 0) then
    Result := true
  else
    Result := false;
end;

procedure MarkRedPixels(Bitmap: TBitmap);
var
  X, Y: Integer;
  Row: PRGBArray;
const
  ColorBlack: TRGBTriple = (rgbtBlue: 0; rgbtGreen: 0; rgbtRed: 0);
  ColorWhite: TRGBTriple = (rgbtBlue: 255; rgbtGreen: 255; rgbtRed: 255);
begin
  for Y := 0 to (Bitmap.Height - 1) do
  begin
    Row := Bitmap.Scanline[Y];
    for X := 0 to (Bitmap.Width - 1) do
    begin
      if IsRedTone(Row[X]) then
        Row[X] := ColorBlack
      else
        Row[X] := ColorWhite;
    end;
    Application.ProcessMessages;
  end;
end;

procedure MarkDarkPixels(Bitmap: TBitmap);
var
  X, Y: Integer;
  Row: PRGBArray;
const
  ColorBlack: TRGBTriple = (rgbtBlue: 0; rgbtGreen: 0; rgbtRed: 0);
  ColorWhite: TRGBTriple = (rgbtBlue: 255; rgbtGreen: 255; rgbtRed: 255);
begin
  for Y := 0 to (Bitmap.Height - 1) do
  begin
    Row := Bitmap.Scanline[Y];
    for X := 0 to (Bitmap.Width - 1) do
    begin
      if IsDarkTone(Row[X]) then
        Row[X] := ColorWhite
      else
        Row[X] := ColorBlack;
    end;
    Application.ProcessMessages;
  end;
end;

{ *** Circle finding code *** }

// Hough transform for recognition of circles
// based on Damir Arh's "Time Recognition" program
// http://www2.arnes.si/~gkrjes12/timerecognition.html
function HoughCircles(Bitmap: TBitmap): TBallData;
const
  MinRadius = 2;
  MaxRadius = 50;
  ColorBlack: TRGBTriple = (rgbtBlue: 0; rgbtGreen: 0; rgbtRed: 0);
  ColorBlue: TRGBTriple = (rgbtBlue: 255; rgbtGreen: 0; rgbtRed: 0);
var
  Parameters: Array[0..SmallWidth-1, 0..SmallHeight-1, MinRadius..MaxRadius] of Byte;
  X, Y: Integer;
  X0, Y0, R: Integer;
  M: Integer;
  Row: PRGBArray;
  I: Integer;
begin
  Application.ProcessMessages;

  // initialize array
  for X0 := 0 to SmallWidth - 1 do
    for Y0 := 0 to SmallHeight - 1 do
      for R := MinRadius to MaxRadius do
        Parameters[X0, Y0, R] := 0;

  // find circles within the list of candidates
  for I := 0 to CandidateList.Count - 1 do
  begin
    for Y := TCandidate(CandidateList.Items[I]).MinY to TCandidate(CandidateList.Items[I]).MaxY do
    begin
      Row := Bitmap.ScanLine[Y];
      for X := TCandidate(CandidateList.Items[I]).MinX to TCandidate(CandidateList.Items[I]).MaxX do
      begin
        if IsBlack(Row[X]) = true then
        begin
          for X0 := TCandidate(CandidateList.Items[I]).MinX + TCandidate(CandidateList.Items[I]).MinR to TCandidate(CandidateList.Items[I]).MaxX - TCandidate(CandidateList.Items[I]).MinR do
          begin
            for Y0 := TCandidate(CandidateList.Items[I]).MinY + TCandidate(CandidateList.Items[I]).MinR to TCandidate(CandidateList.Items[I]).MaxY - TCandidate(CandidateList.Items[I]).MinR do
            begin
              R := Trunc(Sqrt(Power(X - X0, 2) + Power(Y - Y0, 2)));
              if (R >= TCandidate(CandidateList.Items[I]).MinR) and (R <= TCandidate(CandidateList.Items[I]).MaxR) then
                if Parameters[X0, Y0, R] <= 255 then
                  Parameters[X0, Y0, R] := Parameters[X0, Y0, R] + 1;
            end;
          end;
        end;
      end;
      Application.ProcessMessages;
    end;
  end;

  // search for maximum value
  M := -1;
  Result.Radius := -1;
  Result.XPos := -1;
  Result.YPos := -1;
  for R := MinRadius to MaxRadius do
    for X0 := 0 to SmallWidth - 1 do
      for Y0 := 0 to SmallHeight - 1 do
      	if (Parameters[X0, Y0, R] > M) and (Parameters[X0, Y0, R] > 0) then
        begin
          M := Parameters[X0, Y0, R];
          Result.Radius := R;
          Result.XPos := X0;
          Result.YPos := Y0;
        end;

  // usually the radius of the found circle is 2 pixels smaller than the original
  if Result.Radius > 0 then
    inc(Result.Radius, 2);
end;

// search the image for candidates for a filled circle
procedure FillCandidateList(ImgData: TImageData);
var
  X, Y: Integer;
  Pixels: Array[0..SmallWidth - 1, 0..SmallHeight - 1] of Byte;
  PixelCount: Integer;
  AreaExt: TRect;
  Candidate: TCandidate;

  procedure GetExtent(var AreaExt: TRect; var PixelCount: Integer);
  var
    X, Y: Integer;
  begin
    PixelCount := 0;
    AreaExt.Left := 0;
    AreaExt.Right := 0;
    AreaExt.Top := 0;
    AreaExt.Bottom := 0;

    for Y := 0 to SmallHeight - 1 do
    begin
      for X := 0 to SmallWidth - 1 do
      begin
        if Pixels[X, Y] = 1 then
        begin
          if PixelCount = 0 then
          begin
            AreaExt.Left := X;
            AreaExt.Right := X;
            AreaExt.Top := Y;
            AreaExt.Bottom := Y;
          end;
          if (X < AreaExt.Left) then
            AreaExt.Left := X;
          if (X > AreaExt.Right) then
            AreaExt.Right := X;
          if (Y < AreaExt.Top) then
            AreaExt.Top := Y;
          if (Y > AreaExt.Bottom) then
            AreaExt.Bottom := Y;
          Pixels[X, Y] := 2;
          inc(PixelCount);
        end;
      end;
    end;
  end;

  procedure AreaFill(X, Y: Integer);
  begin
    if (Pixels[X, Y] = 0) and (IsDarkTone(GetRGBTriple(ImgData, Y, X)) = True) then
    begin
      Pixels[X, Y] := 1;
      if (X > 0) and (Y > 0) then
        AreaFill(X-1, Y-1);
      if X > 0 then
        AreaFill(X-1, Y);
      if (X > 0) and (Y < SmallHeight - 1) then
        AreaFill(X-1, Y+1);
      if Y > 0 then
        AreaFill(X, Y-1);
      if Y < SmallHeight - 1 then
        AreaFill(X, Y+1);
      if (X < SmallWidth - 1) and (Y > 0) then
        AreaFill(X+1, Y-1);
      if X < SmallWidth - 1 then
        AreaFill(X+1, Y);
      if (X < SmallWidth - 1) and (Y < SmallHeight - 1) then
        AreaFill(X+1, Y+1);
    end;
  end;

begin
  for Y := 0 to SmallHeight - 1 do
    for X := 0 to SmallWidth - 1 do
      Pixels[X, Y] := 0;

  for Y := 0 to SmallHeight - 1 do
  begin
    for X := 0 to SmallWidth - 1 do
    begin
      if (Pixels[X, Y] = 0) and (IsDarkTone(GetRGBTriple(ImgData, Y, X)) = true) then
      begin
        AreaFill(X, Y);
        GetExtent(AreaExt, PixelCount);

        if (PixelCount > 16)
          // at most 30% difference between height and width allowed
          and (Abs((AreaExt.Right - AreaExt.Left) - (AreaExt.Bottom - AreaExt.Top)) <= (AreaExt.Bottom - AreaExt.Top) * 0.3)
          and (Abs((AreaExt.Right - AreaExt.Left) - (AreaExt.Bottom - AreaExt.Top)) <= (AreaExt.Right - AreaExt.Left) * 0.3) then
        begin
          // pixel coverage should be sufficient
          if PixelCount > Round(0.6 * 3.14 * Power((AreaExt.Right - AreaExt.Left + AreaExt.Bottom - AreaExt.Top) / 4, 2)) then
          begin
            // add to candidate list
            Candidate := TCandidate.Create;

            Candidate.MinX := AreaExt.Left - Round((AreaExt.Right - AreaExt.Left) * 0.2);
            if Candidate.MinX < 0 then
              Candidate.MinX := 0;
            Candidate.MaxX := AreaExt.Right + Round((AreaExt.Right - AreaExt.Left) * 0.2);
            if Candidate.MaxX > SmallWidth - 1 then
              Candidate.MaxX := SmallWidth - 1;
            Candidate.MinY := AreaExt.Top - Round((AreaExt.Bottom - AreaExt.Top) * 0.2);
            if Candidate.MinY < 0 then
              Candidate.MinY := 0;
            Candidate.MaxY := AreaExt.Bottom + Round((AreaExt.Bottom - AreaExt.Top) * 0.2);
            if Candidate.MaxY > SmallHeight - 1 then
              Candidate.MaxY := SmallHeight - 1;
            Candidate.MinR := Round(((AreaExt.Right - AreaExt.Left) + (AreaExt.Bottom - AreaExt.Top)) / 4 * 0.8);
            if Candidate.MinR < 2 then
              Candidate.MinR := 2;
            Candidate.MaxR := Round(((AreaExt.Right - AreaExt.Left) + (AreaExt.Bottom - AreaExt.Top)) / 4 * 1.2);
            if Candidate.MaxR > 50 then
              Candidate.MaxR := 50;

            CandidateList.Add(Candidate);
          end;
        end;
      end;
    end;
    Application.ProcessMessages;
  end;
end;

{ *** Other functions *** }

procedure GetImageData(SrcBitmap: TBitmap; var DestImgData: TImageData);
var
  I: Integer;
begin
  Application.ProcessMessages;

  GetMem(DestImgData.Rows, SrcBitmap.Height * SizeOf(PRGBArray));

  for I := 0 to (SrcBitmap.Height - 1) do
    DestImgData.Rows[I] := SrcBitmap.Scanline[I];

  DestImgData.Width := SrcBitmap.Width;
  DestImgData.Height := SrcBitmap.Height;
end;

function FindRedBall(Bitmap: TBitmap): TBallData;
var
  Bitmap1, Bitmap2: TBitmap;
  SrcImgData: TImageData;
  I: Integer;
begin
  Result.XPos := -1;
  Result.YPos := -1;
  Result.Radius := -1;
  if (Bitmap.Width <> LargeWidth) or (Bitmap.Height <> LargeHeight) then
    Exit;

  CandidateList := TList.Create;

  Bitmap1 := TBitmap.Create;
  Bitmap1.Width := Bitmap.Width;
  Bitmap1.Height := Bitmap.Height;
  Bitmap1.PixelFormat := pf24Bit;
  Bitmap1.Canvas.Draw(0, 0, Bitmap);

  MarkRedPixels(Bitmap1);

  Resample(Bitmap1, SmallWidth, SmallHeight);

  GetImageData(Bitmap1, SrcImgData);

  FillCandidateList(SrcImgData);

  Bitmap2 := TBitmap.Create;
  Bitmap2.Width := SrcImgData.Width;
  Bitmap2.Height := SrcImgData.Height;
  Bitmap2.PixelFormat := pf24bit;

  EdgeFilter(SrcImgData, Bitmap2);

  FreeMem(SrcImgData.Rows);
  Bitmap1.Free;

  MarkDarkPixels(Bitmap2);

  Result := HoughCircles(Bitmap2);
  // scale BallData back to the original dimensions
  Result.XPos := Trunc((LargeWidth / SmallWidth) * Result.XPos);
  Result.YPos := Trunc((LargeHeight / SmallHeight) * Result.YPos);
  Result.Radius := Trunc((LargeWidth / SmallWidth) * Result.Radius);

  Bitmap2.Free;

  for I := 0 to CandidateList.Count - 1 do
    TCandidate(CandidateList.Items[I]).Free;
  CandidateList.Free;
end;

end.
