{*******************************************************}
{                                                       }
{  RCXFunctions                                         }
{  Copyright (c) 2003 by Andreas Wulf                   }
{                                                       }
{*******************************************************}
{                                                       }
{  Several high-level functions for controlling an      }
{  RCX unit.                                            }
{                                                       }
{  Last modified: Oct 19, 2003                          }
{                                                       }
{*******************************************************}

unit RCXFunctions;

interface

uses
  RCXBasics;

type
  TRCXIdentifier =
    (RCXMotorA, RCXMotorB, RCXMotorC,
    RCXBackwards, RCXReverse, RCXForwards);

function IsAlive: Boolean;
function PlaySystemSound(Sound: Byte): Boolean;
function StartMotors(Motors: Array of TRCXIdentifier): Boolean;
function StopMotors(Motors: Array of TRCXIdentifier): Boolean;
function SetMotorsPower(Motors: Array of TRCXIdentifier; Power: Byte): Boolean;
function SetMotorDirection(Motor, Direction: TRCXIdentifier): Boolean;
function FlashMotors(Motors, Directions: Array of TRCXIdentifier): Boolean;
function TaskDownload(Number: Integer; Task: String): Boolean;
function GetBatteryLevel: Integer;

implementation

function IsAlive: Boolean;
begin
  Result := SendPacket(CreatePacket(#$10)) <> '';
end;

function PlaySystemSound(Sound: Byte): Boolean;
begin
  Result := SendPacket(CreatePacket(#$51 + Char(Sound))) <> '';
end;

function StartMotors(Motors: Array of TRCXIdentifier): Boolean;
var
  I, Operand: Byte;
begin
  Result := False;

  Operand := $80;
  for I := Low(Motors) to High(Motors) do
  begin
    case Motors[I] of
      RCXMotorA: Operand := Operand or $01;
      RCXMotorB: Operand := Operand or $02;
      RCXMotorC: Operand := Operand or $04;
    else
      exit;
    end
  end;

  Result := SendPacket(CreatePacket(#$21 + Char(Operand))) <> '';
end;

function StopMotors(Motors: Array of TRCXIdentifier): Boolean;
var
  I, Operand: Byte;
begin
  Result := False;

  Operand := $40;
  for I := Low(Motors) to High(Motors) do
  begin
    case Motors[I] of
      RCXMotorA: Operand := Operand or $01;
      RCXMotorB: Operand := Operand or $02;
      RCXMotorC: Operand := Operand or $04;
    else
      exit;
    end
  end;

  Result := SendPacket(CreatePacket(#$21 + Char(Operand))) <> '';
end;

function SetMotorsPower(Motors: Array of TRCXIdentifier; Power: Byte): Boolean;
const
  MotorList: Array [0..2] of Byte = ($01, $02, $04);
var
  I, Operand: Byte;
begin
  Result := False;

  Operand := 0;
  for I := Low(Motors) to High(Motors) do
  begin
    case Motors[I] of
      RCXMotorA: Operand := Operand or $01;
      RCXMotorB: Operand := Operand or $02;
      RCXMotorC: Operand := Operand or $04;
    else
      exit;
    end;
  end;
  if Power < 0 then
    Power := 0
  else
  if Power > 7 then
    Power := 7;

  Result := SendPacket(CreatePacket(#$13 + Char(Operand) + #$02 + Char(Power))) <> '';
end;

function SetMotorDirection(Motor, Direction: TRCXIdentifier): Boolean;
var
  Operand: Byte;
begin
  Result := False;

  case Motor of
    RCXMotorA: Operand := $01;
    RCXMotorB: Operand := $02;
    RCXMotorC: Operand := $04;
  else
    exit;
  end;
  case Direction of
    RCXBackwards: Operand := Operand or $00;
    RCXReverse: Operand := Operand or $40;
    RCXForwards: Operand := Operand or $80;
  else
    exit;
  end;

  Result := SendPacket(CreatePacket(#$E1 + Char(Operand))) <> '';
end;

function FlashMotors(Motors, Directions: Array of TRCXIdentifier): Boolean;
var
  I, OperandLo, OperandHi: Byte;
begin
  Result := False;

  if (Low(Motors) <> Low(Motors)) or (High(Directions) <> High(Directions)) then
    Exit;

  OperandLo := $00;
  OperandHi := $00;
  for I := Low(Motors) to High(Motors) do
  begin
    case Motors[I] of
      RCXMotorA:
        begin
          case Directions[I] of
            RCXForwards: OperandHi := OperandHi or $08;
            RCXBackwards: OperandHi := OperandHi or $40;
          end;
        end;
      RCXMotorB:
        begin
          case Directions[I] of
            RCXForwards: OperandHi := OperandHi or $10;
            RCXBackwards: OperandHi := OperandHi or $80;
          end;
        end;
      RCXMotorC:
        begin
          case Directions[I] of
            RCXForwards: OperandHi := OperandHi or $20;
            RCXBackwards: OperandLo := OperandLo or $01;
          end;
        end;
    else
      Exit;
    end;
  end;

  Result := SendPacket(CreatePacket(#$D2 + Char(OperandLo) + Char(OperandHi))) <> '';
end;

function TaskDownload(Number: Integer; Task: String): Boolean;
var
  Msg, Reply: String;
  I, I2, BlockNo, TransferSize, CheckSum: Integer;
begin
  Result := False;

  // "BeginOfTask"
  Msg := #$25 + #$00 + Char(Number) + #$00 + Char(Lo(Length(Task))) + Char(Hi(Length(Task)));
  Reply := SendPacket(CreatePacket(Msg));
  if Copy(Reply, 6, 1) <> #$00 then
    exit;

  BlockNo := 1;
  TransferSize := 20;
  I := 1;
  Msg := '';
  while I <= Length(Task) do
  begin
    // last block has always number 0
    if I + TransferSize > Length(Task) then
    begin
      BlockNo := 0;
      TransferSize := Length(Task) - I + 1;
    end;

    Msg := #$45 + Chr(Lo(BlockNo)) + Chr(Hi(BlockNo)) + Chr(Lo(TransferSize)) + Chr(Hi(TransferSize));

    CheckSum := 0;
    for I2 := 1 to TransferSize do
    begin
      Msg := Msg + Task[I + I2 - 1];
      CheckSum := CheckSum + Ord(Task[I + I2 - 1]);
    end;

    Msg := Msg + Chr(Lo(CheckSum));

    // transfer block
    Reply := SendPacket(CreatePacket(Msg));
    If Copy(Reply, 6, 1) <> #$00 then
      exit;

    I := I + TransferSize;
    Inc(BlockNo);
  end;

  Result := True;
end;

function GetBatteryLevel: Integer;
var
  Reply: String;
begin
  Result := -1;

  Reply := SendPacket(CreatePacket(#$30));
  if length(Reply) < 8 then
    exit;

  Result := Ord(Reply[6]) + Ord(Reply[8]) * $FF;
end;

end.
