unit WebcamCapture;

interface

uses
  Windows, Forms, Classes, Graphics, VfW, YUVConverts;

type
  TCaptureBitmapEvent = procedure(Bitmap: TBitmap) of Object;

  TCaptureInfo = record
    CaptureWidth: Integer;
    CaptureHeight: Integer;
    CaptureBitsPerPixel: Integer;
  end;

  TWebcamCapture = class
  private
    FCapWnd: HWnd;                   // AVICapture window handle
    FDriverCaps: TCAPDRIVERCAPS;     // driver capabilities
    FCaptureBIHSize: Integer;        // size of capture bmp format info
    FCaptureBIH: PBitmapInfoHeader;  // capture bmp format info
    FStreamBIHSize: Integer;         // size of stream bmp format info
    FStreamBIH: PBitmapInfoHeader;   // stream bmp format info
    FFrameStream: TMemoryStream;     // stream for read a frame
    FFramePixelPointer: PByte;       // pointer to pixel data in FFrameStream
    FVideoCodec: TVideoCodec;        // video codec
    FFrameBitmap: TBitmap;           // bitmap containing the last frame

    FOnCaptureBitmap: TCaptureBitmapEvent;

    procedure CreateBmpHeader(var Bmih: TBitmapInfoHeader; AWidth, AHeight,
      ABitsPerPixel: Integer);
    procedure FrameCallback(hCapWnd: HWND; lpVHDR: PVideoHdr);
    procedure ClearCallback(ACapWnd: HWnd);
    procedure VideoFormatChanged;
    function GetCaptureInfo: TCaptureInfo;
  public
    constructor Create(Width, Height: Integer; ParentHandle: HWnd);
    destructor Destroy; override;

    property OnCaptureBitmap: TCaptureBitmapEvent read FOnCaptureBitmap write FOnCaptureBitmap;
    property DriverCaps: TCAPDRIVERCAPS read FDriverCaps;
    property CaptureInfo: TCaptureInfo read GetCaptureInfo;
    procedure CaptureBitmap;
    procedure DlgVideoFormat;
    procedure DlgVideoSource;
    procedure DlgVideoDisplay;
  end;

implementation

function FrameCallback(hCapWnd: HWnd; lpVHDR: PVideoHdr): DWord; stdcall; far;
var
  OwnerForm: TWebcamCapture;
begin
  OwnerForm := TWebcamCapture(capGetUserData(hCapWnd));
  if Assigned(OwnerForm) then
    OwnerForm.FrameCallBack(hCapWnd, lpVHDR);
  Result := 1;
end;

procedure SetCallback(ACapWnd: HWnd);
begin
  if ACapWnd > 0 then
    capSetCallbackOnFrame(ACapWnd, @FrameCallback);
end;

constructor TWebcamCapture.Create(Width, Height: Integer; ParentHandle: HWnd);
begin
  inherited Create;

  FFrameStream := TMemoryStream.Create;

  FCaptureBIHSize := 0;
  FCaptureBIH := Nil;

  // bitmap info header for frame grab
  FStreamBIHSize := SizeOf(TBitmapInfoHeader);
  GetMem(FStreamBIH, FStreamBIHSize);

  // create capture window
  FCapWnd := capCreateCaptureWindow('Capture Window', WS_CHILD or WS_VISIBLE, 0,
    0, Width, Height, ParentHandle, 0);
  if FCapWnd <> 0 then
  begin
    // the second parameter is the driver index
    if capDriverConnect(FCapWnd, 0) then
    begin
      // get capture driver's capabilities
      capDriverGetCaps(FCapWnd, @FDriverCaps, SizeOf(FDriverCaps));

      // create bitmap header (Width x Height pixels, 24 bit color depth)
      CreateBmpHeader(FStreamBIH^, Width, Height, 24);
      capSetVideoFormat(FCapWnd, FStreamBIH, FStreamBIHSize);
      VideoFormatChanged;

      // set callback
      if capSetUserData(FCapWnd, LongInt(Self)) then
        SetCallback(FCapWnd);
    end
    else
    begin  // unable to conntect to to driver
      DestroyWindow(FCapWnd);
      FCapWnd := 0;
    end;
  end;
  // create frame bitmap
  FFrameBitmap := TBitmap.Create;
end;

destructor TWebcamCapture.Destroy;
begin
  if FCapWnd <> 0 then begin
    capCaptureStop(FCapWnd);
    ClearCallback(FCapWnd);
    capDriverDisconnect(FCapWnd);
    DestroyWindow(FCapWnd);
    FCapWnd := 0;
  end;
  FFramePixelPointer := Nil;
  FFrameStream.Free;
  FFrameBitmap.Free;
  if (FStreamBIHSize > 0) and (FStreamBIH <> Nil) then
    FreeMem(FStreamBIH, FStreamBIHSize);
  if (FCaptureBIHSize > 0) and (FCaptureBIH <> Nil) then
    FreeMem(FCaptureBIH, FCaptureBIHSize);

  inherited Destroy;
end;

procedure TWebcamCapture.CaptureBitmap;
begin
  if FCapWnd = 0 then
    Exit;

  capGrabFrame(FCapWnd);
end;

procedure TWebcamCapture.CreateBmpHeader(var Bmih: TBitmapInfoHeader; AWidth,
  AHeight, ABitsPerPixel: Integer);
begin
  with Bmih do begin
    biSize := SizeOf(TBitmapInfoHeader);
    biWidth := AWidth;
    biHeight := AHeight;
    biPlanes := 1;
    biBitCount := ABitsPerPixel;
    biCompression := BI_RGB;
    biSizeImage := (((biWidth * biBitCount) + 31) div 32) * 4 * biHeight * biPlanes;
    biXPelsPerMeter := 0;
    biYPelsPerMeter := 0;
    biClrUsed := 0;
    biClrImportant := 0;
  end;
end;

procedure TWebcamCapture.FrameCallback(hCapWnd: HWND; lpVHDR: PVideoHdr);
begin
  if hCapWnd = FCapWnd then
  begin
    if Assigned(FFramePixelPointer) then
    begin
      case FVideoCodec of
        vcUnknown: Exit;
        vcRGB: Move(lpVHDR^.lpData^, FFramePixelPointer^, FCaptureBIH^.biSizeImage);
      else
        if Not ConvertCodecToRGB(FVideoCodec, lpVHDR^.lpData, FFramePixelPointer, FCaptureBIH^.biWidth, FCaptureBIH^.biHeight) then Exit;
      end;
      FFrameStream.Position := 0;
      FFrameBitmap.LoadFromStream(FFrameStream);
      if Assigned(FOnCaptureBitmap) then
        FOnCaptureBitmap(FFrameBitmap);
    end;
  end;
end;

procedure TWebcamCapture.ClearCallback(ACapWnd: HWnd);
begin
  if ACapWnd > 0 then
    capSetCallbackOnFrame(ACapWnd, Nil);
end;

procedure TWebcamCapture.VideoFormatChanged;
var
  BmpFileHeader : TBitmapFileHeader;
begin
  if FCapWnd = 0 then Exit;

  // read the really capture format
  FCaptureBIHSize := capGetVideoFormatSize(FCapWnd);
  GetMem(FCaptureBIH, FCaptureBIHSize);
  capGetVideoFormat(FCapWnd, FCaptureBIH, FCaptureBIHSize);

  FVideoCodec := BICompressionToVideoCodec(FCaptureBIH^.biCompression);

  case FVideoCodec of
    vcRGB : begin  // RGB format
      // Copy header
      if FStreamBIHSize <> FCaptureBIHSize then
      begin
        FStreamBIHSize := FCaptureBIHSize;
        ReAllocMem(FStreamBIH, FStreamBIHSize);
      end;
      Move(FCaptureBIH^, FStreamBIH^, FStreamBIHSize);
    end;
    vcYUY2..High(TVideoCodec) : begin
      // 32 bit
      CreateBmpHeader(FStreamBIH^, FCaptureBIH^.biWidth, FCaptureBIH^.biHeight, 32);
    end;
  end;

  // prepare stream for frame
  FFrameStream.Size := SizeOf(BmpFileHeader) + FStreamBIHSize
    + Integer(FStreamBIH^.biSizeImage);
  // fill the header
  FillChar(BmpFileHeader, SizeOf(BmpFileHeader), 0);
  with BmpFileHeader do begin
    bfType := $4D42;  // 'BM'
    bfSize := FFrameStream.Size;
    bfOffBits := SizeOf(BmpFileHeader) + FStreamBIHSize;
  end;
  FFrameStream.Position := 0;
  FFrameStream.WriteBuffer(BmpFileHeader, SizeOf(BmpFileHeader));
  FFrameStream.WriteBuffer(FStreamBIH^, FStreamBIHSize);
  FFramePixelPointer := FFrameStream.Memory;
  Inc(FFramePixelPointer, BmpFileHeader.bfOffBits);
end;

function TWebcamCapture.GetCaptureInfo: TCaptureInfo;
begin
  if Assigned(FCaptureBIH) then
  begin
    Result.CaptureWidth := FCaptureBIH^.biWidth;
    Result.CaptureHeight := FCaptureBIH^.biHeight;
    Result.CaptureBitsPerPixel := FCaptureBIH^.biBitCount;
  end;
end;

procedure TWebcamCapture.DlgVideoFormat;
begin
  if (FDriverCaps.fHasDlgVideoSource = True) and (FCapWnd > 0) then
  begin
    capDlgVideoFormat(FCapWnd);
    VideoFormatChanged;
  end;
end;

procedure TWebcamCapture.DlgVideoSource;
begin
  if (FDriverCaps.fHasDlgVideoFormat = True) and (FCapWnd > 0) then
    capDlgVideoSource(FCapWnd);
end;

procedure TWebcamCapture.DlgVideoDisplay;
begin
  if (FDriverCaps.fHasDlgVideoDisplay = True) and (FCapWnd > 0) then
    capDlgVideoDisplay(FCapWnd);
end;

end.
